from functools import wraps
from typing import Dict, List, Union
from fastapi import Request
from abs_exception_core.exceptions import PermissionDeniedError
from .service import RBACService

def rbac_require_permission(permissions: Union[str, List[str]]):
    """
    Decorator to enforce all required permissions for a user.

    Args:
        permissions (str | list[str]): One or more "resource:action" strings.

    Raises:
        PermissionDeniedError: If the user lacks any one of the required permissions.
    """
    if isinstance(permissions, str):
        permissions = [permissions]

    def decorator(func):
        @wraps(func)
        async def wrapper(
            request:Request,
            *args,rbac_service:RBACService, **kwargs,
        ):
            current_user_uuid = request.state.user.uuid
            if not current_user_uuid:
                raise PermissionDeniedError(
                    detail="User not found (missing 'uuid')."
                )
            for perm in permissions:
                try:
                    module, resource, action = perm.split(":")
                except ValueError:
                    raise ValueError(
                        f"Invalid permission format: '{perm}'. Expected 'module:resource:action'."
                    )
                
                has_permission = rbac_service.check_permission(
                    user_uuid=current_user_uuid, resource=resource, action=action,module=module
                )

                if not has_permission:
                    raise PermissionDeniedError(
                        detail=f"Permission denied: {action} on {resource} in {module}"
                    )
            return await func(*args,request=request,rbac_service=rbac_service, **kwargs)
        return wrapper
    return decorator



from sqlalchemy.orm import relationship

from abs_auth_rbac_core.models.rbac_model import RBACBaseModel


class Role(RBACBaseModel):
    """Role model representing user roles in the system"""
    __tablename__ = "gov_roles"

    # Relationships
    permissions = relationship(
        "Permission", secondary="gov_role_permissions", back_populates="roles"
    )
    users = relationship(
        "Users", secondary="gov_user_roles", back_populates="roles", overlaps="user_roles"
    )
    user_roles = relationship(
        "UserRole", back_populates="role", cascade="all, delete-orphan", overlaps="users,roles"
    )

    eagers = ["permissions", "users"]



from .permissions import Permission
from .roles import Role
from .user_role import UserRole
from .user import Users
from .role_permission import RolePermission
from .user_permission import UserPermission

__all__ = ["Permission", "Role", "UserRole", "Users","RolePermission","UserPermission"]


from fastapi import Depends, Request
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
import logging
from typing import Callable, Any

from .jwt_functions import JWTFunctions
from .auth_functions import get_user_by_attribute
from abs_exception_core.exceptions import UnauthorizedError

security = HTTPBearer()
logger = logging.getLogger(__name__)


# Dependency acting like per-route middleware
def auth_middleware(
    db_session: Callable[...,Any],
    jwt_secret_key:str,
    jwt_algorithm:str
):
    """
    This middleware is used for authentication of the user.
    Args:
        db_session: Callable[...,Any]: Session of the SQLAlchemy database engine
        jwt_secret_key: Secret key of the JWT for jwt functions
        jwt_algorithm: Algorithm used for JWT

    Returns:
    """
    async def get_auth(request: Request, token: HTTPAuthorizationCredentials = Depends(security)):
        jwt_functions = JWTFunctions(secret_key=jwt_secret_key,algorithm=jwt_algorithm)
        try:
            if not token or not token.credentials:
                raise UnauthorizedError(detail="Invalid authentication credentials")

            payload = jwt_functions.get_data(token=token.credentials)
            uuid = payload.get("uuid")

            user = get_user_by_attribute(db_session=db_session,attribute="uuid", value=uuid)
            
            if not user:
                logger.error(f"Authentication failed: User with id {uuid} not found")
                raise UnauthorizedError(detail="Authentication failed")

            # Attach user to request state
            request.state.user = user
            return user 

        except Exception as e:
            logger.error(f"Authentication error: {str(e)}", exc_info=True)
            raise UnauthorizedError(detail="Authentication failed")
    return get_auth


from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.orm import relationship
from abs_auth_rbac_core.models.base_model import BaseModel


class UserPermission(BaseModel):
    """Association model between users and permissions"""

    __tablename__ = "gov_user_permissions"

    user_uuid = Column(String(36), ForeignKey("gov_users.uuid"), index=True)
    permission_uuid = Column(String(36), ForeignKey("gov_permissions.uuid"), index=True)

    user = relationship("Users", back_populates="user_permissions")
    permission = relationship("Permission", back_populates="user_permissions")


from datetime import datetime, timedelta, UTC
from typing import Dict, Type, Callable,Any

import jwt
from fastapi import Depends
from fastapi.security import HTTPBearer
from passlib.context import CryptContext
from abs_exception_core.exceptions import UnauthorizedError,ValidationError


# === JWT Setup ===
pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
bearer_scheme = HTTPBearer()

# === Password Hashing ===

class JWTFunctions:
    def __init__(self,secret_key: str,algorithm: str,expire_minutes: int=None):
        """
        Args:
            secret_key (str): The secret key for the JWT token.
            algorithm (str): The algorithm for the JWT token.
            expire_minutes (int): The expiration time for the JWT token in minutes.
        """
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.expire_minutes = expire_minutes

    def verify_password(self,plain_password: str, hashed_password: str) -> bool:
        """
        Verify a plain password against a hashed password using the password hashing context.

        Args:
            plain_password (str): The plain password to verify.
            hashed_password (str): The hashed password to verify against.
        
        Returns:
            bool: True if the password is verified, False otherwise.
        """
        return pwd_context.verify(plain_password, hashed_password)

    def get_password_hash(self,password: str) -> str:
        """
        Generate a hashed password using the password hashing context.

        Args:
            password (str): The plain password to hash.
        
        Returns:
            str: The hashed password.
        """
        return pwd_context.hash(password)


    # === Token Dependencies ===

    async def get_token(self,token=Depends(bearer_scheme)) -> str:
        """
        Get the token from the bearer scheme.

        Args:
            token (str): The token to get.
        
        Returns:
            str: The token without the bearer prefix.
        """
        return str(token.credentials)

    # === JWT Token Creation & Decoding ===

    def create_jwt_token(self,data: dict, expires_delta: timedelta=None) -> str:
        """
        Create a JWT token.

        Args:
            data (dict): The data to encode in the token.
            expires_delta (timedelta): The expiration time for the token.
        
        Returns:
            str: The JWT token.
        """
        payload = data.copy()
        if expires_delta:
            expire = datetime.now(UTC) + expires_delta
        else:
            expire = datetime.now(UTC) + timedelta(
                minutes=self.expire_minutes
            )
        payload.update({"exp": expire})
        return jwt.encode(payload, self.secret_key, algorithm=self.algorithm)

    def decode_jwt(self,token: str) -> dict:
        """
        Decode a JWT token.

        Args:
            token (str): The token to decode.
        
        Returns:
            dict: The decoded token.
        """
        try:
            return jwt.decode(token, self.secret_key, algorithms=[self.algorithm])
        except jwt.ExpiredSignatureError:
            raise UnauthorizedError(detail="Token has expired")
        except jwt.InvalidTokenError:
            raise  ValidationError(detail="Invalid token")

    def get_data(self,token: str = Depends(get_token)) -> Dict:
        """
        Get the data from the JWT token.

        Args:
            token (str): The token to get the data from.
        
        Returns:
            Dict: The decoded token.
        """
        return self.decode_jwt(token)
    
    # === Token Generators ===

    def create_access_token(self,data: Dict, expires_delta: timedelta=None) -> str:
        """
        Create an access token.

        Args:
            data (Dict): The data to encode in the token.
            expires_delta (timedelta): The expiration time for the token.
        
        Returns:
            str: The access token.
        """
        return self.create_jwt_token(data=data, expires_delta=expires_delta)